package org.jastadd.ros2rag.compiler;

import beaver.Parser;
import org.jastadd.ros2rag.ast.*;
import org.jastadd.ros2rag.compiler.options.CommandLine;
import org.jastadd.ros2rag.compiler.options.CommandLine.CommandLineException;
import org.jastadd.ros2rag.compiler.options.Option;
import org.jastadd.ros2rag.compiler.options.StringOption;
import org.jastadd.ros2rag.parser.Ros2RagParser;
import org.jastadd.ros2rag.scanner.Ros2RagScanner;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;

public class Compiler {

  private StringOption optionOutputDir;
  private StringOption optionInputGrammar;
  private StringOption optionInputRos2Rag;

  private ArrayList<Option<?>> options;
  private CommandLine commandLine;

  public Compiler() {
    options = new ArrayList<>();
    addOptions();
  }

  public int run(String[] args) throws CommandLineException, CompilerException {
    options = new ArrayList<>();
    addOptions();
    commandLine = new CommandLine(options);
    commandLine.parse(args);

    String outputDir = "gen"; // should not be used, but otherwise there is a compiler warning
    if (optionOutputDir.isSet()) {
      outputDir = optionOutputDir.getValue();
    } else {
      System.out.println("No output output dir is set. Assuming '" + outputDir + "'.");
    }

    printMessage("Running ROS2RAG Preprocessor");

    if (!optionInputGrammar.isSet()) {
      return error("specify an input grammar");
    }

    if (!optionInputRos2Rag.isSet()) {
      return error("specify the ros2rag definition file");
    }

    List<String> otherArgs = commandLine.getArguments();
    if (!otherArgs.isEmpty()) {
      printMessage("Superfluous arguments will be ignored: " + otherArgs);
    }
    Ros2Rag ros2Rag = parseProgram(optionInputGrammar.getValue(), optionInputRos2Rag.getValue());

    printMessage("Writing output files");
    writeToFile(outputDir + "/Grammar.relast", ros2Rag.getProgram().generateAbstractGrammar());
    writeToFile(outputDir + "/ROS2RAG.jadd", ros2Rag.generateAspect());
    return 0;
  }


  public static void main(String[] args) {
    try {
      new Compiler().run(args);
    } catch (CommandLineException | CompilerException e) {
      System.err.println(e.getMessage());
      System.exit(-1);
    }
  }

  private void printMessage(String message) {
    System.out.println(message);
  }

  private void writeToFile(String filename, String str) throws CompilerException {
    try {
      PrintWriter writer = new PrintWriter(filename);
      writer.print(str);
      writer.close();
    } catch (Exception e) {
      throw new CompilerException("Could not write to file " + filename, e);
    }
  }

  private void addOptions() {
    optionOutputDir = addOption(new StringOption("outputDir", "target directory for the generated files."));
    optionInputGrammar = addOption(new StringOption("inputGrammar", "base grammar."));
    optionInputRos2Rag = addOption(new StringOption("inputRos2Rag", "ros2rag definition file."));
  }

  private <OptionType extends Option<?>> OptionType addOption(OptionType option) {
    options.add(option);
    return option;
  }

  private Ros2Rag parseProgram(String inputGrammarFileName, String inputRos2RagFileName) throws CompilerException {
    Program program = new Program();
    Ros2Rag ros2Rag;

    try (BufferedReader reader = Files.newBufferedReader(Paths.get(inputGrammarFileName))) {
      Ros2RagScanner scanner = new Ros2RagScanner(reader);
      Ros2RagParser parser = new Ros2RagParser();
      GrammarFile inputGrammar = (GrammarFile) parser.parse(scanner);
      inputGrammar.dumpTree(System.out);
      program.addGrammarFile(inputGrammar);
      inputGrammar.treeResolveAll();
    } catch (IOException | Parser.Exception e) {
      throw new CompilerException("Could not parse grammar file " + inputGrammarFileName, e);
    }

    try (BufferedReader reader = Files.newBufferedReader(Paths.get(inputRos2RagFileName))) {
      Ros2RagScanner scanner = new Ros2RagScanner(reader);
      Ros2RagParser parser = new Ros2RagParser();
      ros2Rag = (Ros2Rag) parser.parse(scanner, Ros2RagParser.AltGoals.ros2rag);
      ros2Rag.setProgram(program);
    } catch (IOException | Parser.Exception e) {
      throw new CompilerException("Could not parse ros2rag file " + inputRos2RagFileName, e);
    }

    ros2Rag.dumpTree(System.out);
    ros2Rag.treeResolveAll();
    ros2Rag.dumpTree(System.out);

    return ros2Rag;
  }

  private void parse(Program program, Reader reader, String file) {
    Ros2RagScanner scanner = new Ros2RagScanner(reader);
    Ros2RagParser parser = new Ros2RagParser();

    try {
      GrammarFile newFile = (GrammarFile) parser.parse(scanner);
      program.addGrammarFile(newFile);
    } catch (IOException e) {
      error(e.getMessage());
    } catch (Parser.Exception e) {
      System.err.println("Parse error in file " + file);
      System.err.println(e.getMessage());
      System.exit(1);
    }
  }

  protected int error(String message) {
    System.err.println("Error: " + message);
    System.err.println();
    System.err.println("Usage: java -jar ros2rag.jar [--option1] [--option2=value] ...  <filename1> <filename2> ... ");
    System.err.println("Options:");
    System.err.print(commandLine.printOptionHelp());
    return 1;
  }

  public static class CompilerException extends Exception {
    CompilerException(String message, Throwable cause) {
      super(message, cause);
    }
  }
}

